#include "disk_info.h"

; ===============================
#define via_portb   $0300
#define via_ddrb    $0302
#define via_ddra    $0303
#define via_t1cl    $0304
#define via_t1ch    $0305
#define via_t1ll    $0306
#define via_t1lh    $0307
#define via_t2ll    $0308
#define via_t2ch    $0309
#define via_sr      $030a
#define via_acr     $030b
#define via_pcr     $030c
#define via_ifr     $030d
#define via_ier     $030e
#define via_porta   $030f

; ===============================
    .zero
    *=loader_zp_start

retry_counter       .dsb 1      ; Number of attempts at loading data (ie: not quite clear what happens when this fails...
sectors_to_go       .dsb 1      ; How many sectors do we still need to load for this file
current_track       .dsb 1      ; Index of the track being loaded
current_sector      .dsb 1      ; Index of the sector being loaded
current_side        .dsb 1      ; Has the bit 4 set to 0 or 1 to be used as a mask on the Microdisc control register (other bits have to be set to zero)
irq_save_a          .dsb 1      ; To preserve the accumulator value in the IRQ code

; Unpack test
ptr_source          .dsb 2      ; Packed source data adress
ptr_destination     .dsb 2      ; Destination adress where we depack
ptr_destination_end .dsb 2      ; Point on the end of the depacked stuff
ptr_source_back     .dsb 2      ; Temporary used to hold a pointer on depacked stuff
offset              .dsb 2
mask_value          .dsb 1
nb_dst              .dsb 1

; ===============================
; This is called from the bootsectors
    .text
    *=location_loader

    jmp StartUp

; ===============================
; This values are set in bootsectors code
fdc_offset
    .byt 0
cmd_writesector
    .byt 0
cmd_readsector
    .byt 0
cmd_seek
    .byt 0
cmd_forceinterrupt
    .byt 0
fdc_flag_discside
    .byt 0

; ===============================
; This file is generated by the floppy builder
#define LOADER
#include "floppy_description.h"

; ===============================
StartUp
    ; for any case disable interrupts
    ; actually they are disabled in 
    ; microdisc/jasmin boot sector
    sei
    ; reset the stack
    ldx #$ff
    txs
    cld

    jsr init_via_defaults
        
#ifdef LOADER_HIRES
    ;
    ; Switch to HIRES
    ;
    ldy #39             ; From $9900 to $c000 is 39 pages (9984 bytes)
    lda #0
loop_hires_outer
    tax
loop_hires_inner
__auto_hires
    sta $9900,x
    inx
    bne loop_hires_inner
    inc __auto_hires+2
    dey
    bne loop_hires_outer

    lda #30             ; Write hires switch
    sta $bfdf
#endif

    ; Load the main program
    ldx #LOADER_MAIN
    jsr LoadData
    
    ; Give control to the application and hope it knows what to do
__auto_execute_address
    jmp $feed

; ===============================
init_via_defaults
    ; this code is from Atmos ROM
    lda #$ff
    sta via_ddra
    lda #$f7
    sta via_ddrb
    lda #$b7
    sta via_portb
    lda #$dd
    sta via_pcr
    lda #$7f
    sta via_ier
    lda #$40 
    sta via_acr
    lda #$c0
    sta via_ier
    lda #$10
    sta via_t1ll
    sta via_t1cl
    lda #$27
    sta via_t1lh
    sta via_t1ch
    rts

; ===============================
irq_def_handler
    sta irq_save_a
    pla
    pha
    and #%00010000  ; Check the saved B flag to detect a BRK
    beq from_via

from_brk
    ; TODO: indicate brk irq
    bne from_brk

from_via
    lda via_ifr
    and #$40
    beq from_irq
    ; Clear IRQ event
    bit via_t1cl

    lda irq_save_a
    jmp irq_do_nothing

from_irq
    lda irq_save_a

irq_do_nothing
    rti

; ===============================
; X=File index
; A=Low
; Y=High
SetLoadAddress
    sta FileLoadAdressLow,x
    tya
    sta FileLoadAdressHigh,x
    rts


; ===============================
; X=File index
SaveData
    ; Init operation params
    jsr Start_Operation

    sei
;
; Loop to write all the sectors
;
write_sectors_loop
    ;
    ldy #0
w_loop_copy

__auto_write_address
    lda $c000,y         ; Load the byte from original location
    sta loader_sector_buffer,y    ; Store it to page 2

    iny
    bne w_loop_copy

    nop
    nop

    ;
    jsr write_next_sector

    ; Next sector
    inc __auto_write_address+2
    dec sectors_to_go
    bne write_sectors_loop

    lda cmd_forceinterrupt 
    ldy fdc_offset 
    sta fdc_command_register,y 
    lda fdc_data,y 
    cli
    
    rts

write_next_sector
    ;
    ldy fdc_offset

    ; Check if we have reached the end of the track
    lda current_sector
    cmp #FLOPPY_SECTOR_PER_TRACK+1
    bne w_same_track

    ; Next track
    inc current_track
    lda current_track
    cmp #FLOPPY_TRACK_NUMBER
    bne w_stay_on_same_side

    ; Reset to the first track on the other side
    lda #0
    sta current_track

    lda fdc_flag_discside
    sta current_side
w_stay_on_same_side

    ; Reset the sector position
    lda #1
    sta current_sector

w_same_track
    jsr WaitCommand

    lda current_sector
    sta fdc_sector_register,y
    inc current_sector

    ; Check if the drive is on the correct track
    lda current_track
    cmp fdc_track_register,y
    beq w_stay_on_the_track

    ; Set the new track
    sta fdc_data,y

    lda cmd_seek
    sta fdc_command_register,y
    jsr WaitCompletion

w_stay_on_the_track
    lda #fdc_flags_mask
    ora current_side
    sta fdc_flags,y

    lda cmd_writesector
    sta fdc_command_register,y

    jsr WaitCommand

    ;
    ; Write the sector data
    ;
    ldx #0
write_byte
    lda fdc_drq,y
    bmi write_byte

    lda loader_sector_buffer,x      ; Load byte from page 2
    sta fdc_data,y
    inx
    bne write_byte

    lda fdc_status_register,y
    and #$1C
    
    jsr WaitCompletion
    rts

; ===============================
; X=File index
LoadData
    ; We have to start somewhere no matter what, compressed or not
    jsr Start_Operation

    sei
    
    ; Now at this stage we have to check if the data is compressed or not
    lda FileStoredSizeLow,x
    cmp FileSizeLow,x
    bne load_compressed_data

    lda FileStoredSizeHigh,x
    cmp FileSizeHigh,x
    bne load_compressed_data

;
; Loop to read all the sectors
;
read_sectors_loop
    ;
    jsr read_next_sector

    ldy #0
r_loop_copy
    lda loader_sector_buffer,y    ; Load the byte from page 2

__auto_read_address
    sta $c000,y         ; Store it to the final location
    iny
    bne r_loop_copy

    ; Next sector
    inc __auto_read_address+2
    dec sectors_to_go
    bne read_sectors_loop

    lda cmd_forceinterrupt 
    ldy fdc_offset 
    sta fdc_command_register,y 
    lda fdc_data,y 
    cli
    
    rts

load_compressed_data
    clc
    lda FileLoadAdressLow,x
    sta ptr_destination+0
    adc FileSizeLow,x
    sta ptr_destination_end+0

    lda FileLoadAdressHigh,x
    sta ptr_destination+1
    adc FileSizeHigh,x
    sta ptr_destination_end+1

    jsr unpack_data

    lda cmd_forceinterrupt 
    ldy fdc_offset 
    sta fdc_command_register,y 
    lda fdc_data,y 
    cli
    
    rts

read_next_sector
    ldy fdc_offset

    ; Check if we have reached the end of the track
    lda current_sector
    cmp #FLOPPY_SECTOR_PER_TRACK+1
    bne r_same_track

    ; Next track
    inc current_track
    lda current_track
    cmp #FLOPPY_TRACK_NUMBER
    bne r_stay_on_same_side

    ; Reset to the first track on the other side
    lda #0
    sta current_track

    lda fdc_flag_discside
    sta current_side
r_stay_on_same_side

    ; Reset the sector position
    lda #1
    sta current_sector

r_same_track
    jsr WaitCommand

    lda current_sector
    sta fdc_sector_register,y
    inc current_sector

    ; Check if the drive is on the correct track
    lda current_track
    cmp fdc_track_register,y
    beq r_stay_on_the_track

    ; Set the new track
    sta fdc_data,y

    lda cmd_seek
    sta fdc_command_register,y
    jsr WaitCompletion

r_stay_on_the_track
    lda #fdc_flags_mask
    ora current_side
    sta fdc_flags,y

    lda cmd_readsector
    sta fdc_command_register,y

    jsr WaitCommand

    ;
    ; Read the sector data
    ;
    ldx #0
read_byte
    lda fdc_drq,y
    bmi read_byte

    lda fdc_data,y
    sta loader_sector_buffer,x      ; Store the byte in page 2
    inx
    bne read_byte

    lda fdc_status_register,y
    and #$1C

    jsr WaitCompletion
    rts

Start_Operation
    ldy fdc_offset

    ; Make sure the microdisc IRQ is disabled
    jsr WaitCompletion

    lda #fdc_flags_mask          ; Disable the FDC (Eprom select + FDC Interrupt request)
    sta fdc_flags,y

    jsr WaitCompletion

    ;
    ; Setup, we use the table to find out where the file is located on the floppy,
    ; where to write it in memory, and how large it is.
    ;
    lda FileLoadAdressHigh,x
    sta __auto_execute_address+2
    sta __auto_write_address+2
    sta __auto_read_address+2

    lda FileLoadAdressLow,x
    sta __auto_execute_address+1
    sta __auto_write_address+1
    sta __auto_read_address+1

    ; Starting track
    ldy #%00000000              ; Side 0
    lda FileStartTrack,x        ; If the track id is larger than 128, it means it is on the other side of the floppy
    bpl first_side
    ; The file starts on the second side
    ldy fdc_flag_discside       ; Side 1
    and #%01111111              ; Mask out the extra bit
first_side
    sty current_side
    sta current_track

    ldy fdc_offset

    ; First sector
    lda FileStartSector,x
    sta current_sector

    ; FileSizeLow/FileSizeHigh
    ; Number of sectors to load
    lda FileStoredSizeHigh,x
    sta sectors_to_go
    lda FileStoredSizeLow,x
    beq sectors_to_go_calc
    inc sectors_to_go
sectors_to_go_calc
    rts


WaitCommand
    ldx #wait_status_floppy
waitcommand
    nop
    nop
    dex
    bne waitcommand
    rts

WaitCompletion
    txa
    pha

    php
    nop
    nop
    nop

    ldx #4
r_wait_completion
    dex
    bne r_wait_completion
    plp

r2_wait_completion
    lda fdc_status_register,y
    lsr
    bcs r2_wait_completion

    ;asl
    pla
    tax
    rts

GetNextByte
    php
    lda __fetchByte_lo
    bne __fetchByte
    nop
    nop
    nop
    nop
    jsr read_next_sector
    nop
    nop
    nop
    nop
    ldx #0
    ldy #0
__fetchByte
__fetchByte_lo = (*+1)
    lda loader_sector_buffer
    inc __fetchByte_lo
    plp
    rts


; void file_unpack(void *ptr_dst,void *ptr_src)

; Need to be called with valid values in:
; ptr_destination
; ptr_source
; ptr_destination_end (Destination + size)
unpack_data
    ; Initialise variables
    ; We try to keep "y" null during all the code,
    ; so the block copy routine has to be sure that
    ; y is null on exit
    ldy #0
    sty __fetchByte_lo
    lda #1
    sta mask_value

unpack_loop
    ; Handle bit mask
    lsr mask_value
    bne end_reload_mask

    jsr GetNextByte     ; Read from source stream

    ror
    sta mask_value
end_reload_mask
    bcc back_copy

    ; Copy one byte from the source stream
    jsr GetNextByte     ; Read from source stream
    sta (ptr_destination),y

    lda #1
    sta nb_dst

unpack_end_loop
    ; We increase the current destination pointer,
    ; by a given value, white checking if we reach
    ; the end of the buffer.
    clc
    lda ptr_destination
    adc nb_dst
    sta ptr_destination

    bcc skip_unp
    inc ptr_destination+1
skip_unp

    cmp ptr_destination_end
    lda ptr_destination+1
    sbc ptr_destination_end+1
    bcc unpack_loop
    rts

back_copy
    ; Copy a number of bytes from the already unpacked stream
    ; Here we know that y is null. So no need for clearing it.
    ; Just be sure it`s still null at the end.
    ; At this point, the source pointer points to a two byte
    ; value that actually contains a 4 bits counter,
    ; and a 12 bit offset to point back into the depacked stream.
    ; The counter is in the 4 high order bits.
    ;clc  <== No need, since we access this routie from a BCC
    jsr GetNextByte     ; Read from source stream
    adc #1
    sta offset
    jsr GetNextByte     ; Read from source stream
    tax
    and #$0f
    adc #0
    sta offset+1

    txa
    lsr
    lsr
    lsr
    lsr
    clc
    adc #3
    sta nb_dst

    sec
    lda ptr_destination
    sbc offset
    sta ptr_source_back
    lda ptr_destination+1
    sbc offset+1
    sta ptr_source_back+1

    ; Beware, in that loop, the direction is important
    ; since RLE like depacking is done by recopying the
    ; very same byte just copied... Do not make it a
    ; reverse loop to achieve some speed gain...

copy_loop
    lda (ptr_source_back),y ; Read from already unpacked stream
    sta (ptr_destination),y ; Write to destination buffer
    iny
    cpy nb_dst
    bne copy_loop

    ldy #0
    beq unpack_end_loop
    rts

_EndLoaderCode

;
; This is free memory that can be used, when it reaches zero then the loader start address should be changed
;

    .dsb $FFF1 - _EndLoaderCode

_Vectors

#if ( _Vectors - $FFF1 )
#error - Vector address is incorrect, loader will crash
#else

;
; Here are the functions that the user can call from his own application
;
_ApiSetLoadAddress  .byt $4c,<SetLoadAddress,>SetLoadAddress    ; $FFF1
_ApiSetLoadData     .byt $4c,<LoadData,>LoadData                ; $FFF4
_ApiSetSaveData     .byt $4c,<SaveData,>SaveData                ; $FFF7

;
; These three HAVE to be at these precise adresses, they map to hardware registers
;
_VectorNMI          .word irq_do_nothing                        ; FFFA-FFFB - NMI Vector (Usually points to $0247)
_VectorReset        .word irq_do_nothing                        ; FFFC-FFFD - RESET Vector (Usually points to $F88F)
_VectorIRQ          .word irq_def_handler                       ; FFFE-FFFF - IRQ Vector (Normally points to $0244)

#echo Remaining space in the loader code:
#print (_Vectors - _EndLoaderCode)
#endif

; End of the loader - Nothing should come after because it`s out of the addressable memory range :)
